const fs = require('fs'),
      sys = require('sys'),
      events = require('events');

/**
 * Constructor
 */
function FakeDB(path) {

  // call the parent
  events.EventEmitter.call(this);

  var self = this;

  this._hash = {};
  this._list = [];
  this._count = 0;
  this._operations = [];
  this._logging = false;
  this._flushing = false;

  this.path = path;
  
  this._rstr = fs.createReadStream(this.path, {
    flags: 'r',
    encoding: 'utf-8',
    mode: 0666
  });
  
  this._wstr = fs.createWriteStream(this.path, {
    flags: 'a+',
    encoding: 'utf-8',
    mode: 0666
  });

  this._wstr.on('drain', function () {
    self._flushing = false;
    if (self._operations.length) {
      // if we have more stuff to write
      self._flush();
    } else {
      self.emit('drain');
    }
  });

  this._load();
}

/**
 * We'll be emitting 'load', 'drain' and 'error'
 */
sys.inherits(FakeDB, events.EventEmitter);

/**
 * Adds operation to log
 */
FakeDB.prototype._log = function (entry) {
  if (this._logging) {
    this._operations.push(entry);
  }
};

/**
 * Load data from a db file
 */
FakeDB.prototype._load = function () {

  var buffer = '',
      stream = null,
      self = this;
      
  this._rstr.on('error', function (err) {
    if (err.code == 'ENOENT') {
      self._logging = true;
      self.emit('load');
    }
  });

  this._rstr.on('data', function (chunk) {
    buffer += chunk;
  });

  this._rstr.on('end', function () {

    var lines = buffer.split('\n');
    lines.forEach(function (row) {
      if (row.length) {
        try {
          row = JSON.parse(row);
          if ('key' in row && 'value' in row) {
            self.set(row.key, row.value);
          } else {
            throw new Error('Invalid record encountered');
          }
        } catch (e) {
          self.emit('error', e);
        }
      }
    });

    // log all operations after loading data
    self._logging = true;

    self.emit('load');
  });

};

/**
 * Flush the log queue into the db file
 */
FakeDB.prototype._flush = function () {

  if (this._flushing || !this._operations.length) {
    return;
  }

  this._flushing = true;

  var data = null,
      self = this;

  while (data = this._operations.shift()) {
    this._wstr.write(JSON.stringify(data) + '\n', function (err) {
      if (err) {
        self.emit('error', new Error('Failed to append to log file.'));
      }
    });
  }
};

/**
 * Set a key in the database
 */
FakeDB.prototype.set = function (key, value) {
  
  if (!(key in this._hash)) {
    this._list.push(key);
  }
  
  this._hash[key] = value;
  
  if (this._logging) {
    this._log({ key: key, value: value });
    this._flush();
  }
};

/**
 * Remove an object
 */
FakeDB.prototype.remove = function (key) {
  if (key in this._hash) {
    delete this._list.splice(this._list.indexOf(key), 1);
    delete this._hash[key];
    if (this._logging) {
      this._log({ key: key, value: null });
      this._flush();
    }
  }
};

/**
 * Get a stored key
 */
FakeDB.prototype.get = function (key, cb) {
  if (key in this._hash) {
    cb.call(undefined, key, this._hash[key]);
  }
};

/**
 * Get all records
 */
FakeDB.prototype.all = function (cb, opt) {

  opt = opt || {
    offset: 0,
    limit: 0
  };

  var i = (opt.offset > this._list.length ? this._list.length : opt.offset),
      n = (opt.limit > 0 ? opt.limit : this._list.length),
      k = v = null;

  while (i < n) {
    k = this._list[i++];
    v = this._hash[k];
    cb.call(undefined, k, v);
  }
};

/**
 * First record
 */
FakeDB.prototype.first = function (cb) {
  if (this._list.length) {
    cb.call(undefined, this._list[0], this._hash[this._list[0]]);
  }
};

/**
 * Last record
 */
FakeDB.prototype.last = function (cb) {
  if (this._list.length) {
    var k = this._list[this._list.length - 1];
    var v = this._hash[k];
    cb.call(undefined, k, v);
  }
};


/**
 * Returns number of records in DB
 */
FakeDB.prototype.count = function () {
  return this._list.length;
};

exports.init = function (path) {
  return new FakeDB(path);
};
